import functools
import os
from json import dumps

import click
from httpx import HTTPStatusError, RequestError

from . import deploy
from . import log as deployment_log
from .deploy import load_vars
from .http_client import create_client
from .load_module import load_module

help_footer = "environment defaults to $LUSID_ENV and token defaults to $FBN_ACCESS_TOKEN"


# decorator which applies a set of common options to the command
def common_options(fn):
    options = [
        click.option(
            "-e",
            "--environment",
            envvar="LUSID_ENV",
            required=True,
            help="Base url to the LUSID domain, eg https://foo.lusid.com. "
            "Defaults to env var $LUSID_ENV",
        ),
        click.option(
            "-t",
            "--access-token",
            type=str,
            envvar="FBN_ACCESS_TOKEN",
            required=True,
            help="A LUSID access token. Defaults to env var $FBN_ACCESS_TOKEN",
        ),
    ]
    to_wrap = fn
    for wrapper in options:
        wrapped = wrapper(to_wrap)
        to_wrap = wrapped
    return to_wrap


def http_exception_handler(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except RequestError as exc:
            msg = (
                f"An error occurred while requesting {exc.request.url!r}"
                f"before we got a response - {exc}"
            )
            raise click.ClickException(msg)
        except HTTPStatusError as exc:
            status = exc.response.status_code
            base_msg = " ".join(
                [
                    f"The server responded with {exc.response.status_code}",
                    f"while requesting {exc.request.method} {exc.request.url!r}",
                ]
            )
            if exc.response.is_server_error:
                detail = [exc.response.text]
            elif status == 401:
                detail = [
                    "Unauthorized. "
                    "This may indicate that the provided access token is invalid or expired."
                ]
            elif exc.response.json().get("name", "") == "AccessDenied":
                detail = [exc.response.json().get("detail"), exc.response.json().get("instance")]
            elif status == 400:
                detail = [exc.response.text]
            elif status == 404:
                error_detail = exc.response.json().get("detail", None)
                detail = [error_detail if error_detail else exc.response.text]
            else:
                detail = [exc.response.text]
            msg_str = "\n".join([base_msg] + detail)
            raise click.ClickException(msg_str)

    return wrapper


def runtime_error_handler(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except RuntimeError as exc:
            raise click.ClickException(str(exc))

    return wrapper


@click.group()
@click.version_option(message="%(version)s")
def cli():
    pass  # base cli class


@cli.command(epilog=help_footer)
@common_options
@http_exception_handler
@runtime_error_handler
def setup(environment, access_token):
    """Perform initial setup of the environment"""
    client = create_client(environment, access_token)
    deployment_log.setup(client)


@click.group()
def log():
    """Commands for interacting with the deployment log"""
    pass


@log.command(name="list", epilog=help_footer)
@common_options
@click.argument("deployment", type=click.STRING, default=None, required=0)
@http_exception_handler
@runtime_error_handler
def list_deployments(deployment, environment, access_token):
    """List deployments, or resources within a specified deployment.
    If no deployment id passed, returns all deployment ids ordered by last modified"""
    client = create_client(environment, access_token)
    if deployment:
        for line in deployment_log.list_resources_for_deployment(client, deployment_id=deployment):
            click.echo(deployment_log.format_log_line(line))
    else:
        for line in deployment_log.list_deployments(client):
            click.echo(f"   * {line}")


@log.command(epilog=help_footer)
@common_options
@click.argument("deployment", type=click.STRING)
@click.argument("resource_id", type=click.STRING)
@http_exception_handler
@runtime_error_handler
def get(environment, access_token, deployment, resource_id):
    """Get the logged state of a resource."""
    client = create_client(environment, access_token)
    for r in deployment_log.get_resource(client, deployment, resource_id):
        click.echo(f"    {r.resource_type}")
        click.echo(dumps(r.state.__dict__, indent=4))


@log.command(epilog=help_footer)
@common_options
@click.argument("deployment", type=click.STRING)
@click.argument("resource_id", type=click.STRING)
@http_exception_handler
@runtime_error_handler
def deps(deployment, resource_id, environment, access_token):
    """Show dependencies for this resource."""
    client = create_client(environment, access_token)
    (index, dependencies) = deployment_log.get_dependencies_map(client, deployment)
    deployment_log.print_tree(index, dependencies, resource_id, "=>")


@log.command(epilog=help_footer)
@common_options
@click.argument("deployment", type=click.STRING)
@click.argument("resource_id", type=click.STRING)
@http_exception_handler
@runtime_error_handler
def uses(deployment, resource_id, environment, access_token):
    """Show resources dependent on this resource."""
    client = create_client(environment, access_token)
    (index, rdeps) = deployment_log.get_dependents_map(client, deployment)
    deployment_log.print_tree(index, rdeps, resource_id, "<=")


@log.command(epilog=help_footer)
@common_options
@click.option(
    "-f",
    "--force",
    is_flag=True,
    required=False,
    default=False,
    show_default=True,
    help="Remove the entry even if there are dependencies on given resource id",
)
@click.option(
    "-r",
    "--recursive",
    is_flag=True,
    required=False,
    default=False,
    show_default=True,
    help="Remove the entry and any resources that  depend on it",
)
@click.argument("deployment", type=click.STRING)
@click.argument("resource_id", type=click.STRING)
@click.pass_context
@http_exception_handler
@runtime_error_handler
def rm(ctx, deployment, resource_id, force, recursive, environment, access_token):
    """Remove resources from a deployment."""
    client = create_client(environment, access_token)
    resource_log = deployment_log.get_resource(client, deployment, resource_id)
    if len(resource_log) == 0:
        ctx.fail(f"Resource {resource_id} not found in deployment {deployment}")
    _, dependents = deployment_log.get_dependents_map(client, deployment)
    resource_uses = dependents.get(resource_id, None)
    if resource_uses and force is False:
        ctx.fail(f"   Cannot remove entry for '{resource_id}' as {resource_uses} depends on it")
    elif recursive is True and resource_uses:
        for id in resource_uses:
            deployment_log.remove(client, deployment, id)
        deployment_log.remove(client, deployment, resource_id)
        click.echo(
            f"   Removed entry for '{resource_id}' in deployment '{deployment}' "
            f"and its dependencies {dependents}"
        )
    else:
        deployment_log.remove(client, deployment, resource_id)
        click.echo(f"   Removed entry for '{resource_id}' in deployment '{deployment}'")


cli.add_command(log)


@cli.command(epilog=help_footer)
@common_options
@click.option("-v", "--vars_file", type=click.File("r"))
@click.argument("script_path", type=click.Path(readable=True))
@http_exception_handler
@runtime_error_handler
def run(script_path, environment, vars_file, access_token):
    """Deploy the configuration

    Loads the configuration script from SCRIPT_PATH
    and runs against lusid.
    vars_file will be injected into the configure function
    """
    host_vars = load_vars(vars_file)
    try:
        module = load_module(script_path, os.path.basename(script_path))
        if getattr(module, "configure", None) is None:
            raise click.ClickException(
                f"Failed importing user config: [{script_path}]. Error : No configure function found"
            )

        d = module.configure(host_vars)
    except ImportError as e:
        raise click.ClickException(f"Failed importing user config: [{script_path}]. Error : \n{e}")
    deploy(d, environment, access_token)


if __name__ == "__main__":
    cli()



from __future__ import annotations

from typing import Any, Dict

import httpx
from pydantic import BaseModel, Field

from .resource_abc import CamelAlias, Resource


class SequenceResource(CamelAlias, BaseModel, Resource):
    id: str = Field(exclude=True)
    scope: str
    code: str
    increment: int | None = None
    min_value: int | None = None
    max_value: int | None = None
    start: int | None = None
    cycle: bool | None = None
    pattern: str | None = None

    def read(self, client, old_state) -> Dict[str, Any]:
        return client.request("get", f"/api/api/sequences/{self.scope}/{self.code}").json()

    def create(self, client: httpx.Client):
        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True)
        client.request("POST", f"/api/api/sequences/{self.scope}", json=desired)
        return {"scope": self.scope, "code": self.code}

    def update(self, client: httpx.Client, old_state):
        if [old_state.scope, old_state.code] != [self.scope, self.code]:
            raise (RuntimeError("Cannot change the scope/code on a sequence"))
        remote = self.read(client, old_state)
        desired = self.model_dump(
            mode="json", exclude_none=True, exclude={"scope", "code"}, by_alias=True
        )
        effective = remote | desired
        if effective == remote:
            return None
        raise RuntimeError("Cannot modify a sequence")

    @staticmethod
    def delete(client, old_state):
        raise RuntimeError("Cannot delete a sequence")

    def deps(self):
        return []



from abc import ABC, abstractmethod
from types import SimpleNamespace
from typing import Any, Dict, Optional, Sequence, Union

from httpx import Client as httpxClient
from pydantic import AliasGenerator, ConfigDict, alias_generators

alias_generator = ConfigDict(
    alias_generator=AliasGenerator(
        serialization_alias=lambda field_name: alias_generators.to_camel(field_name)
    )
)


class Ref(ABC):
    id: str

    @abstractmethod
    def attach(self, client: httpxClient) -> None:
        pass

    def deps(self):
        return []


class Resource(ABC):
    model_config = alias_generator

    id: str

    @abstractmethod
    def read(self, client: httpxClient, old_state: SimpleNamespace) -> None | Dict[str, Any]:
        pass

    @abstractmethod
    def create(self, client: httpxClient) -> Optional[Dict[str, Any]]:
        pass

    @abstractmethod
    def update(self, client: httpxClient, old_state) -> Union[None, Dict[str, Any]]:
        pass

    @staticmethod
    @abstractmethod
    def delete(client: httpxClient, old_state) -> None:
        pass

    @abstractmethod
    def deps(self) -> Sequence["Resource|Ref"]:
        pass


class CamelAlias(ABC):
    model_config = alias_generator



from __future__ import annotations

import copy
from enum import Enum
from typing import Any, Dict, List, Optional, Sequence

import httpx
from pydantic import BaseModel, Field, field_serializer

from .resource_abc import CamelAlias, Ref, Resource


class ActionId(BaseModel):
    """ActionId resource used with IdSelector

    Example
    -------
    >>> from fbnconfig.access import ActionId
    >>> ActionId(scope="myscope", activity="execute", entity="Feature")

    -------
    """

    scope: str
    activity: str
    entity: str


class MatchAllSelector(BaseModel):
    type_name: str = Field("matchAllSelectorDefinition", init=False, exclude=True)
    actions: List[ActionId]
    name: Optional[str] = None
    description: Optional[str] = None


class IdSelector(BaseModel):
    """IdSelector resource used with PolicyResource

    Example
    -------
    >>> from fbnconfig.access import IdSelector, ActionId
    >>> IdSelector(
            name="feature_id_selector",
            description="feature_id_selector",
            identifier={"scope": "myscope", "code": "mycode"},
            actions=[ActionId(scope="myscope", activity="execute", entity="Feature")])
    """

    type_name: str = Field("idSelectorDefinition", init=False, exclude=True)
    identifier: Dict[str, str]
    actions: List[ActionId]
    name: Optional[str] = None
    description: Optional[str] = None


class MetadataExpression(CamelAlias, BaseModel):
    metadata_key: str
    operator: str
    text_value: str | None


class MetadataSelector(BaseModel):
    type_name: str = Field("metadataSelectorDefinition", init=False, exclude=True)
    actions: List[ActionId]
    name: str | None = None
    description: str | None = None
    expressions: List[MetadataExpression]


class PolicySelector(CamelAlias, BaseModel):
    type_name: str = Field("policySelectorDefinition", init=False, exclude=True)
    actions: List[ActionId]
    name: str | None = None
    description: str | None = None
    identity_restriction: Dict[str, str] | None = None
    restriction_selectors: Sequence[Selector] | None = None

    @field_serializer("restriction_selectors", when_used="always")
    def serialize_selectors(self, selectors: Any):
        # convert array of selectors
        return [{selector.type_name: selector} for selector in selectors]


Selector = IdSelector | MatchAllSelector | MetadataSelector | PolicySelector


class WhenSpec(BaseModel):
    """
    WhenSpec resource used with PolicyResource

    Example
    -------
    >>> from fbnconfig.access import WhenSpec
    >>>WhenSpec(activate="2024-08-31T18:00:00.0000000+00:00")

    Notes
    -------
    When deactivate is not supplied, the policy is valid from time in activate until end of time
    """

    activate: str
    deactivate: Optional[str] = None


class Grant(str, Enum):
    """Type of grant used with PolicyResource

    Available values are: Allow, Deny and Undefined
    """

    ALLOW = "Allow"
    DENY = "Deny"
    UNDEFINED = "Undefined"


class PolicyRef(BaseModel, Ref):
    id: str = Field(exclude=True, init=True)
    code: str
    scope: str = "default"

    def attach(self, client):
        try:
            client.request("get", f"/access/api/policies/{self.code}", params={"scope": self.scope})
        except httpx.HTTPStatusError as ex:
            if ex.response.status_code == 404:
                raise RuntimeError(f"Policy {self.scope}/{self.code} not found")
            else:
                raise ex


class PolicyResource(BaseModel, Resource):
    """Manage a policy

    Attributes
    -------
    id: str
        Resource identifier
    """

    id: str = Field(exclude=True)
    code: str
    scope: str = Field("default", init=False, exclude=True)
    description: str
    applications: List[str]
    grant: Grant
    selectors: Sequence[Selector]
    when: WhenSpec

    @field_serializer("selectors", when_used="always")
    def serialize_selectors(self, selectors: Any):
        # new version: convert list of selectors to list of dicts
        # where the key in each dict is the type of selector
        return [{selector.type_name: selector} for selector in selectors]

    def read(self, client, old_state):
        remote = client.request(
            "get", f"/access/api/policies/{self.code}", params={"scope": self.scope}
        ).json()
        remote.pop("id")
        remote.pop("links")
        return remote

    def create(self, client: httpx.Client):
        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True)
        client.request("POST", "/access/api/policies", json=desired)
        return {"id": self.id, "code": self.code}

    def update(self, client: httpx.Client, old_state):
        if old_state.code != self.code:
            raise (RuntimeError("Cannot change the code on a policy"))
        get = self.read(client, old_state)
        remote = copy.deepcopy(get)
        desired = self.model_dump(
            mode="json", exclude_none=True, exclude={"scope", "code"}, by_alias=True
        )
        if (
            desired["when"].get("deactivate", None) is None
        ):  # deactivate is defaulted on the server so not a difference unless we set it
            remote["when"].pop("deactivate")
        if desired == remote:
            return None
        client.request("put", f"/access/api/policies/{self.code}", json=desired)
        return {"id": self.id, "code": self.code}

    @staticmethod
    def delete(client, old_state):
        client.request("DELETE", f"/access/api/policies/{old_state.code}")

    def deps(self):
        return []


class PolicyCollectionRef(BaseModel, Ref):
    id: str = Field(exclude=True, init=True)
    code: str
    scope: str = "default"

    def attach(self, client):
        try:
            client.get(f"/access/api/policycollections/{self.code}", params={"scope": self.scope})
        except httpx.HTTPStatusError as ex:
            if ex.response.status_code == 404:
                raise RuntimeError(f"PolicyCollection {self.scope}/{self.code} not found")
            else:
                raise ex


class PolicyCollectionResource(CamelAlias, BaseModel, Resource):
    id: str = Field(exclude=True)
    # collections can be referenced by scope, but it can't be specified
    scope: str = Field("default", init=False, exclude=True)
    code: str
    description: str | None = None
    policies: Sequence[PolicyResource | PolicyRef] = []
    policy_collections: Sequence[PolicyCollectionResource | PolicyCollectionRef] | None = []
    metadata: Dict[str, List[str]] | None = None

    @field_serializer("policies", when_used="json")
    def serialize_policies(self, policies: Any):
        return [{"code": p.code, "scope": p.scope} for p in policies]

    @field_serializer("policy_collections", when_used="json")
    def serialize_collections(self, collections: Any):
        return [{"code": p.code, "scope": p.scope} for p in collections]

    def read(self, client, old_state) -> Dict[str, Any]:
        scope = old_state.scope
        code = old_state.code
        params = {"scope": scope} if scope is not None else None
        remote = client.get(f"/access/api/policycollections/{code}", params=params).json()
        remote.pop("links")
        remote.pop("id")
        return remote

    def create(self, client) -> Dict[str, Any]:
        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True)
        res = client.request("POST", "/access/api/policycollections", json=desired)
        return res.json()["id"]

    def update(self, client, old_state) -> Dict[str, Any] | None:
        if old_state.code != self.code:
            self.delete(client, old_state)
            return self.create(client)
        remote = self.read(client, old_state)
        desired = self.model_dump(mode="json", exclude_none=True, exclude={"code"}, by_alias=True)
        if desired == remote:
            return None
        client.request("put", f"/access/api/policycollections/{self.code}", json=desired)
        return {"code": self.code, "scope": old_state.scope}

    @staticmethod
    def delete(client, old_state):
        client.request("DELETE", f"/access/api/policycollections/{old_state.code}")

    def deps(self) -> List[Resource | Ref]:
        deps: List[Resource | Ref] = []
        for pol in self.policies:
            deps.append(pol)
        if self.policy_collections:
            for col in self.policy_collections:
                deps.append(col)
        return deps


class PolicyIdRoleResource(CamelAlias, BaseModel):
    """Used to refer to a policy resource in a role resource"""

    policies: Sequence[PolicyResource | PolicyRef] | None = []
    policy_collections: List[PolicyCollectionResource] | None = []

    @field_serializer("policies", when_used="json")
    def serialize_policy_identifiers(self, policies: List[PolicyResource | PolicyRef]):
        # this takes a policy resource as a dep, but it only needs to send the identifiers
        return [{"code": p.code, "scope": p.scope} for p in policies]

    @field_serializer("policy_collections", when_used="json")
    def serialize_collection_identifiers(self, policies: Any):
        # this takes a policy resource as a dep, but it only needs to send the identifiers
        return [{"code": p.code, "scope": p.scope} for p in policies]


class Permission(str, Enum):
    """Permission type used on a role resource"""

    READ = "Read"
    WRITE = "Write"
    EXECUTE = "Execute"


class NonTransitiveSupervisorRoleResource(BaseModel):
    roles: Sequence[RoleResource | RoleRef]

    @field_serializer("roles", when_used="always")
    def serialize_role_identifiers(self, roles: List[RoleResource | RoleRef]):
        return [{"code": r.code, "scope": r.scope} for r in roles]


class RoleResourceRequest(CamelAlias, BaseModel):
    non_transitive_supervisor_role_resource: NonTransitiveSupervisorRoleResource | None = None
    policy_id_role_resource: PolicyIdRoleResource | None = None


class RoleRef(BaseModel, Ref):
    """Reference an existing Role"""

    id: str = Field(exclude=True, init=True)
    scope: str = "default"
    code: str

    def attach(self, client):
        try:
            params = {"scope": self.scope}
            client.request("get", f"/access/api/roles/{self.code}", params=params)
        except httpx.HTTPStatusError as ex:
            if ex.response.status_code == 404:
                raise RuntimeError(f"Role {self.scope}/{self.code} not found")
            else:
                raise ex


class RoleResource(CamelAlias, BaseModel, Resource):
    """Define a role resource"""

    id: str = Field(exclude=True)
    scope: str = Field("default", exclude=True, init=False)
    code: str
    description: str | None = None
    resource: None | RoleResourceRequest = None
    when: WhenSpec
    permission: Permission
    role_hierarchy_index: Optional[int] = None

    def read(self, client, old_state):
        get = client.request("get", f"/access/api/roles/{self.code}")
        remote = get.json()
        remote.pop("id")
        remote.pop("links")
        return remote

    def create(self, client):
        body = self.model_dump(mode="json", exclude_none=True, by_alias=True)
        client.request("POST", "/access/api/roles", json=body)
        return {"id": self.id, "code": self.code}

    def update(self, client: httpx.Client, old_state):
        if old_state.code != self.code:
            raise (RuntimeError("Cannot change the code on a role"))

        remote = self.read(client, old_state)
        desired = self.model_dump(mode="json", exclude={"code"}, exclude_none=True, by_alias=True)
        remote["when"].pop(
            "deactivate"
        )  # deactivate is defaulted on the server so not a difference unless we set it
        remote.pop("roleHierarchyIndex")  # set by the server
        if desired == remote:
            return None
        client.request("put", f"/access/api/roles/{self.code}", json=desired)
        return {"id": self.id, "code": self.code}

    @staticmethod
    def delete(client, old_state):
        client.request("DELETE", f"/access/api/roles/{old_state.code}")

    def deps(self):
        if self.resource is None or self.resource.policy_id_role_resource is None:
            return []
        res = self.resource.policy_id_role_resource
        pol_deps = [v for v in res.policies] if res.policies is not None else []
        col_deps = [v for v in res.policy_collections] if res.policy_collections is not None else []
        return pol_deps + col_deps



from __future__ import annotations

import hashlib
import json
from datetime import datetime
from enum import StrEnum
from types import SimpleNamespace
from typing import Any

import httpx
from httpx import Client as httpxClient
from pydantic import BaseModel, ConfigDict, Field, computed_field, field_serializer, model_serializer

from . import identity, lumi, scheduler
from .resource_abc import CamelAlias, Ref, Resource


class WorkerRef(BaseModel, Ref):
    """Reference an existing worker

    Example
    ----------
    >>> from fbnconfig import workflows
    >>> workflows.WorkerRef(
    ...  id="reconciliation-worker",
    ...  scope="default",
    ...  code="GroupReconciliation")


    Attributes
    ----------
    id : str
         Resource identifier.
    scope : str
        Scope of the worker.
    code: str
        Code of the worker.
    """

    id: str = Field(exclude=True)
    scope: str
    code: str

    def attach(self, client):
        try:
            client.get(f"/workflow/api/workers/{self.scope}/{self.code}")
        except httpx.HTTPStatusError as ex:
            if ex.response.status_code == 404:
                raise RuntimeError(f"Worker {self.scope}/{self.code} does not exist")
            else:
                raise ex


class WorkerResource(BaseModel, Resource):
    """Define a worker

    Example
    ----------
    >>> from fbnconfig import workflows
    >>> workflows.WorkerResource(
    ...  id="myworker",
    ...  scope="default",
    ...  code="MyWorker",
    ...  display_name="DoSomeWork",
    ...  description="New description",
    ...  worker_configuration=workflows.LuminesceView(
    ...    view=lumi.ViewRef(id="my-view", provider="Views.Unit.Something"))
    ... )


    Attributes
    ----------
    id : str
         Resource identifier.
    scope : str
        Scope of the worker.
    code: str
        Code of the worker.
    display_name: str
        Human-readable name for the worker
    description: str | None
        Optional description of the worker
    worker_configuration: Fail | HealthCheck | LuminesceView | SchedulerJob | Sleep
        Worker configuration

    """

    id: str = Field(exclude=True)
    scope: str = Field(exclude=True, init=True)
    code: str = Field(exclude=True, init=True)
    display_name: str
    description: str | None = None
    worker_configuration: Fail | HealthCheck | LuminesceView | SchedulerJob | Sleep

    @computed_field(alias="id")
    def worker_id(self) -> dict[str, str]:
        return {"scope": self.scope, "code": self.code}

    def read(self, client: httpxClient, old_state: SimpleNamespace) -> dict[str, Any] | None:
        scope = old_state.scope
        code = old_state.code
        return client.get(f"/workflow/api/workers/{scope}/{code}").json()

    # noinspection PyProtectedMember
    def __resolve_worker_version(self) -> str:
        if isinstance(self.worker_configuration, SchedulerJob):
            if isinstance(self.worker_configuration.job, scheduler.JobResource):
                dump = self.worker_configuration.job.model_dump(
                    mode="json", exclude_none=True, by_alias=True
                )
                return hashlib.sha256(json.dumps(dump, sort_keys=True).encode()).hexdigest()
            if isinstance(self.worker_configuration.job, scheduler.JobRef):
                # noinspection PyProtectedMember
                version = self.worker_configuration.job._content_hash
                ref_id = f"{self.worker_configuration.job_id}"
                assert version is not None, f"Expected JobRef {ref_id} to be attached"
                return version

        if isinstance(self.worker_configuration, LuminesceView):
            if isinstance(self.worker_configuration.view, lumi.ViewResource):
                return self.worker_configuration.view._get_content_hash()
            if isinstance(self.worker_configuration.view, lumi.ViewRef):
                return self.worker_configuration.view._version or "0"
        return "0"

    def create(self, client: httpxClient) -> dict[str, Any] | None:
        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True)
        client.post("/workflow/api/workers", json=desired)

        return {
            "id": self.id,
            "scope": self.scope,
            "code": self.code,
            "worker_version": self.__resolve_worker_version(),
        }

    def update(self, client: httpxClient, old_state: SimpleNamespace) -> dict[str, Any] | None:
        if [self.scope, self.code] != [old_state.scope, old_state.code]:
            self.delete(client, old_state)
            return self.create(client)

        remote = self.read(client, old_state) or {}
        remote.pop("version")
        remote.pop("parameters")
        remote.pop("resultFields")
        remote.pop("links")
        remote.pop("id")

        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True, exclude={"worker_id"})
        effective = remote | desired
        if desired == effective and old_state.worker_version == self.__resolve_worker_version():
            return None

        client.put(f"/workflow/api/workers/{self.scope}/{self.code}", json=desired).json()
        return {
            "id": self.id,
            "scope": self.scope,
            "code": self.code,
            "worker_version": self.__resolve_worker_version(),
        }

    @staticmethod
    def delete(client: httpxClient, old_state: SimpleNamespace) -> None:
        client.delete(f"/workflow/api/workers/{old_state.scope}/{old_state.code}")

    def deps(self) -> list[Any]:
        if isinstance(self.worker_configuration, LuminesceView) and isinstance(
            self.worker_configuration.view, (lumi.ViewResource, lumi.ViewRef)
        ):
            return [self.worker_configuration.view]
        if isinstance(self.worker_configuration, SchedulerJob):
            return [self.worker_configuration.job]
        return []


class TaskDefinitionRef(BaseModel, Ref):
    """Reference an existing task definition

    Example
    ----------
    >>> from fbnconfig import workflows
    >>> workflows.TaskDefinitionRef(
    ...  id="task-definition-ref",
    ...  scope="DQ",
    ...  code="ExceptionTask")


    Attributes
    ----------
    id : str
         Resource identifier.
    scope : str
        Scope of the task definition.
    code: str
        Code of the task definition.
    """

    id: str = Field(exclude=True)
    scope: str
    code: str

    def attach(self, client):
        try:
            client.get(f"/workflow/api/taskdefinitions/{self.scope}/{self.code}").json()
        except httpx.HTTPStatusError as ex:
            if ex.response.status_code == 404:
                raise RuntimeError(f"Task definition {self.scope}/{self.code} does not exist")
            else:
                raise ex


class TaskDefinitionResource(BaseModel, Resource):
    id: str = Field(exclude=True)
    scope: str = Field(exclude=True)
    code: str = Field(exclude=True)
    display_name: str
    initial_state: InitialState
    states: list[TaskStateDefinition]
    description: str | None = None
    triggers: list[TriggerDefinition] | None = None
    transitions: list[TaskTransitionDefinition] | None = None
    field_schema: list[TaskFieldDefinition] | None = None
    actions: list[ActionDefinition] | None = None

    @computed_field(alias="id")
    def task_def_id(self) -> dict[str, str]:
        return {"scope": self.scope, "code": self.code}

    def read(self, client: httpxClient, old_state: SimpleNamespace) -> None | dict[str, Any]:
        remote = client.get(f"/workflow/api/taskdefinitions/{old_state.scope}/{old_state.code}").json()
        return remote

    def create(self, client: httpxClient) -> dict[str, Any] | None:
        desired = self.model_dump(mode="json", by_alias=True, exclude_none=True)
        source_version = hashlib.sha256(json.dumps(desired, sort_keys=True).encode()).hexdigest()
        remote_version = client.post("/workflow/api/taskdefinitions", json=desired).json()["version"][
            "asAtVersionNumber"
        ]
        return {
            "id": self.id,
            "scope": self.scope,
            "code": self.code,
            "source_version": source_version,
            "remote_version": remote_version,
        }

    def update(self, client: httpxClient, old_state) -> dict[str, Any] | None:
        if self.scope != old_state.scope or self.code != old_state.code:
            self.delete(client, old_state)
            return self.create(client)

        remote = self.read(client, old_state) or {}
        desired = self.model_dump(mode="json", by_alias=True, exclude_none=True)
        remote_version = remote["version"]["asAtVersionNumber"]
        source_version = hashlib.sha256(json.dumps(desired, sort_keys=True).encode()).hexdigest()
        if source_version == old_state.source_version and remote_version == old_state.remote_version:
            return None

        remote_version = client.put(
            f"/workflow/api/taskdefinitions/{self.scope}/{self.code}", json=desired
        ).json()["version"]["asAtVersionNumber"]
        return {
            "id": self.id,
            "scope": self.scope,
            "code": self.code,
            "source_version": source_version,
            "remote_version": remote_version,
        }

    @staticmethod
    def delete(client: httpxClient, old_state) -> None:
        client.delete(f"/workflow/api/taskdefinitions/{old_state.scope}/{old_state.code}")

    def deps(self) -> list[Any]:
        deps = []
        for t in self.transitions if self.transitions else []:
            deps.extend(t.deps())
        for a in self.actions if self.actions else []:
            deps.extend(a.deps())

        return deps


class Fail(BaseModel):
    type: str = "Fail"


class HealthCheck(BaseModel):
    type: str = "HealthCheck"
    url: str


class LuminesceView(BaseModel):
    type: str = "LuminesceView"
    view: lumi.ViewResource | lumi.ViewRef = Field(exclude=True, init=True)

    @computed_field
    def name(self) -> str:
        return self.view.provider


class SchedulerJob(BaseModel):
    type: str = "SchedulerJob"
    job: scheduler.JobRef | scheduler.JobResource = Field(exclude=True, init=True)

    @computed_field(alias="jobId")
    def job_id(self) -> dict[str, str]:
        assert self.job is not None, "Job should not be None"
        return {"scope": self.job.scope, "code": self.job.code}


class Sleep(BaseModel):
    type: str = "Sleep"


class TaskStateDefinition(BaseModel):
    model_config = ConfigDict(use_enum_values=True, frozen=True)
    name: str


class TaskFieldDefinitionType(StrEnum):
    STRING = "String"
    DECIMAL = "Decimal"
    DATETIME = "DateTime"
    BOOLEAN = "Boolean"


class ReadOnlyStateType(StrEnum):
    ALL_STATES = "AllStates"
    INITIAL_STATE = "InitialState"
    TERMINAL_STATE = "TerminalState"
    SELECTED_STATES = "SelectedStates"


class ReadOnlyStates(BaseModel, CamelAlias):
    model_config = ConfigDict(use_enum_values=True, frozen=True)
    state_type: str | ReadOnlyStateType
    selected_states: list[str | TaskStateDefinition] | None = None

    @field_serializer("selected_states")
    def states(self, selected_states) -> list[str] | None:
        return [s.name if isinstance(s, TaskStateDefinition) else s for s in selected_states]

    def __hash__(self) -> int:
        return hash((self.state_type, frozenset(self.selected_states or frozenset())))


class ValueConstraintType(StrEnum):
    SUGGESTED = "Suggested"
    VALIDATED = "Validated"


class ValueSourceType(StrEnum):
    ACCEPTABLE_VALUES = "AcceptableValues"


class ValueConstraints(BaseModel, CamelAlias):
    model_config = ConfigDict(use_enum_values=True, frozen=True)
    constraint_type: str | ValueConstraintType
    value_source_type: str | ValueSourceType
    acceptable_values: list[str]

    def __hash__(self) -> int:
        return hash((self.constraint_type, self.value_source_type, frozenset(self.acceptable_values)))


class TaskFieldDefinition(BaseModel, CamelAlias):
    model_config = ConfigDict(use_enum_values=True, frozen=True)
    name: str
    type: str | TaskFieldDefinitionType
    read_only_states: ReadOnlyStates | None = None
    value_constraints: ValueConstraints | None = None

    @field_serializer("type")
    def type_serializer(self, field_type: str | TaskFieldDefinitionType | None):
        return field_type.name if isinstance(field_type, TaskFieldDefinitionType) else field_type


class InitialState(BaseModel, CamelAlias):
    name: str | TaskStateDefinition
    required_fields: list[str | TaskFieldDefinition] | None = []

    @field_serializer("name")
    def name_serializer(self, name: str | TaskStateDefinition):
        return name.name if isinstance(name, TaskStateDefinition) else name

    @field_serializer("required_fields")
    def req_fields(self, required_fields) -> list[str] | None:
        return [f.name if isinstance(f, TaskFieldDefinition) else f for f in required_fields]


class TriggerDefinition(BaseModel):
    name: str
    type: str

    @model_serializer()
    def serialize_model(self):
        return {"name": self.name, "trigger": {"type": self.type}}


class FieldMapping(BaseModel, CamelAlias):
    map_from: str | TaskFieldDefinition | None = None
    set_to: str | TaskFieldDefinition | None = None

    @field_serializer("*")
    def map_serializer(self, value: str | TaskFieldDefinition | None) -> str | None:
        return value.name if isinstance(value, TaskFieldDefinition) else value


class ChildTaskConfiguration(BaseModel, CamelAlias):
    task_definition: TaskDefinitionResource | TaskDefinitionRef = Field(
        serialization_alias="taskDefinitionId"
    )
    task_definition_as_at: str | datetime | None = None
    initial_trigger: str | TriggerDefinition | None = None
    child_task_fields: dict[str | TaskFieldDefinition, FieldMapping] | None = None
    map_stacking_key_from: str | TaskFieldDefinition | None = None

    @field_serializer("task_definition")
    def task_definition_id(
        self, task_definition: TaskDefinitionResource | TaskDefinitionRef
    ) -> dict[str, str]:
        return {"scope": task_definition.scope, "code": task_definition.code}

    @field_serializer("task_definition_as_at")
    def task_def_as_at(self, task_as_at: str | datetime | None) -> str | None:
        return task_as_at.isoformat() if isinstance(task_as_at, datetime) else task_as_at

    @field_serializer("initial_trigger")
    def trigger(self, initial_trigger: str | TriggerDefinition | None) -> str | None:
        return (
            initial_trigger.name if isinstance(initial_trigger, TriggerDefinition) else initial_trigger
        )

    @field_serializer("map_stacking_key_from")
    def mskf(self, map_stacking_key_from: str | TaskFieldDefinition | None) -> str | None:
        return (
            map_stacking_key_from.name
            if isinstance(map_stacking_key_from, TaskFieldDefinition)
            else map_stacking_key_from
        )

    @field_serializer("child_task_fields")
    def ctf(self, child_task_fields: dict[str | TaskFieldDefinition, FieldMapping]):
        return {
            k.name if isinstance(k, TaskFieldDefinition) else k: v for k, v in child_task_fields.items()
        }


class CreateChildTasksAction(BaseModel, CamelAlias):
    type: str = "CreateChildTasks"
    child_task_configurations: list[ChildTaskConfiguration]

    def deps(self) -> list[TaskDefinitionResource | TaskDefinitionRef]:
        return [c.task_definition for c in self.child_task_configurations]


class WorkerStatusTriggers(BaseModel, CamelAlias):
    started: str | TriggerDefinition | None = None
    completed_with_results: str | TriggerDefinition | None = None
    completed_no_results: str | TriggerDefinition | None = None
    failed_to_start: str | TriggerDefinition | None = None
    failed_to_complete: str | TriggerDefinition | None = None

    @field_serializer("*")
    def serialize_fields(self, value) -> str:
        return value.name if isinstance(value, TriggerDefinition) else value


class TriggerParentTaskAction(BaseModel):
    type: str = Field(default="TriggerParentTask", init=False)
    trigger: str | TriggerDefinition

    @field_serializer("trigger")
    def trigger_serialization(self, trigger: str | TriggerDefinition) -> str:
        return trigger.name if isinstance(trigger, TriggerDefinition) else trigger


class ResultantChildTaskConfiguration(BaseModel, CamelAlias):
    child_task_configuration: ChildTaskConfiguration
    result_matching_pattern: str | None = None

    @model_serializer()
    def serialize_model(self):
        result_pattern = (
            {"resultMatchingPattern": {"filter": self.result_matching_pattern}}
            if self.result_matching_pattern
            else None
        )
        if result_pattern is None:
            return self.child_task_configuration.model_dump(by_alias=True, exclude_none=True)
        else:
            return (
                self.child_task_configuration.model_dump(by_alias=True, exclude_none=True)
                | result_pattern
            )


class RunWorkerAction(BaseModel, CamelAlias):
    type: str = "RunWorker"
    worker: WorkerResource | WorkerRef = Field(exclude=True)
    worker_as_at: str | datetime | None = None
    worker_parameters: dict[str | TaskFieldDefinition, FieldMapping] | None = None
    worker_status_triggers: WorkerStatusTriggers | None = None
    child_task_configurations: list[ResultantChildTaskConfiguration] | None = None
    # in seconds
    worker_timeout: int | None = None

    @computed_field()
    def worker_id(self) -> dict[str, str]:
        return {"scope": self.worker.scope, "code": self.worker.code}

    @field_serializer("worker_as_at")
    def worker_as_at_s(self, worker_as_at: str | datetime | None) -> str | None:
        return worker_as_at.isoformat() if isinstance(worker_as_at, datetime) else worker_as_at

    @field_serializer("worker_parameters")
    def worker_params(self, worker_parameters: dict[str | TaskFieldDefinition, FieldMapping]):
        return {
            k.name if isinstance(k, TaskFieldDefinition) else k: v for k, v in worker_parameters.items()
        }

    def deps(self) -> list[Any]:
        child_tasks = self.child_task_configurations if self.child_task_configurations else []
        return [self.worker] + [c.child_task_configuration.task_definition for c in child_tasks]


class ActionDefinition(BaseModel, CamelAlias):
    name: str
    action_details: CreateChildTasksAction | RunWorkerAction | TriggerParentTaskAction
    run_as_user_id: str | identity.UserResource | identity.UserRef | None = None

    @field_serializer("run_as_user_id")
    def user_id(
        self, run_as_user_id: str | identity.UserResource | identity.UserRef | None
    ) -> str | None:
        return run_as_user_id if isinstance(run_as_user_id, str | None) else run_as_user_id.user_id

    def deps(self) -> list[Any]:
        result = []
        if isinstance(self.run_as_user_id, identity.UserResource | identity.UserRef):
            result.append(self.run_as_user_id)
        if isinstance(self.action_details, CreateChildTasksAction | RunWorkerAction):
            result.extend(self.action_details.deps())
        return result


class TaskTransitionDefinition(BaseModel, CamelAlias):
    from_state: str | TaskStateDefinition
    to_state: str | TaskStateDefinition
    trigger: str | TriggerDefinition
    guard: str | None = None
    action: str | ActionDefinition | None = None

    @field_serializer("from_state", "to_state", "trigger", "action")
    def state_serializer(
        self, value: str | TaskStateDefinition | TriggerDefinition | ActionDefinition
    ) -> str:
        return (
            value.name
            if isinstance(value, (TaskStateDefinition, TriggerDefinition, ActionDefinition))
            else value
        )

    def deps(self):
        return self.action.deps() if isinstance(self.action, ActionDefinition) else []



from __future__ import annotations

import hashlib
import json
import sys
import textwrap
import time
from enum import Enum, StrEnum
from typing import Any, Dict, List

import httpx
from pydantic import BaseModel, Field

from .backoff_handler import BackoffHandler
from .resource_abc import CamelAlias, Ref, Resource


# from the sdk
class TaskStatus(StrEnum):
    CREATED = "Created"
    WAITING_FOR_ACTIVATION = "WaitingForActivation"
    WAITING_TO_RUN = "WaitingToRun"
    RUNNING = "Running"
    WAITING_FOR_CHILDREN_TO_COMPLETE = "WaitingForChildrenToComplete"
    RAN_TO_COMPLETION = "RanToCompletion"
    CANCELED = "Canceled"
    FAULTED = "Faulted"


def start_query(client: httpx.Client, sql: str) -> str:
    return client.put(
        "/honeycomb/api/SqlBackground", content=sql, headers={"Content-type": "text/plain"}
    ).json()["executionId"]


def wait_for_background(client, execution_id) -> bool:
    status = get_status(client, execution_id)
    match status:
        case TaskStatus.FAULTED | TaskStatus.CANCELED:
            raise RuntimeError("Query was " + status + ". Execution id: " + execution_id)
        case (
            TaskStatus.CREATED
            | TaskStatus.WAITING_FOR_ACTIVATION
            | TaskStatus.WAITING_TO_RUN
            | TaskStatus.WAITING_FOR_CHILDREN_TO_COMPLETE
        ):
            return False
        case TaskStatus.RUNNING:
            return False
        case TaskStatus.RAN_TO_COMPLETION:
            return True
        case _:
            raise RuntimeError("Unknown status: " + status + ". Execution id: " + execution_id)


def fetch(client, execution_id):
    res = client.get(f"/honeycomb/api/SqlBackground/{execution_id}/jsonproper")
    return res.json()


def get_status(client: httpx.Client, execution_id: str) -> str:
    """Get the status of the query in Luminesce

    Returns:
        str: string containing the query status value.
    """
    return client.get(f"/honeycomb/api/SqlBackground/{execution_id}").json()["status"]


def query(client: httpx.Client, sql: str, backoff_handler: BackoffHandler) -> Dict[Any, Any]:
    execution_id: str = start_query(client, sql)
    progress = False
    while True:
        if progress is False:
            backoff_handler.sleep()
            progress = wait_for_background(client, execution_id)
        else:
            #  Pause after the last wait_for_background_call, otherwise increased chances of 429s
            backoff_handler.sleep()
            return fetch(client, execution_id)


class ParameterType(str, Enum):
    BigInt = "BigInt"
    Boolean = "Boolean"
    Date = "Date"
    DateTime = "DateTime"
    Decimal = "Decimal"
    Double = "Double"
    Int = "Int"
    Table = "Table"
    Text = "Text"


class VariableType(str, Enum):
    Scalar = "@@"
    Table = "@"


class Variable(BaseModel):
    name: str
    type: VariableType
    sql: str

    def init_str(self):
        #  eg @scalar = select 2 + 2
        return f"{self.type.value}{self.name} = {self.sql}"

    def with_str(self):
        return f"{self.type.value}{self.name}"


class Parameter(BaseModel, CamelAlias):
    name: str
    type: ParameterType
    value: Any
    set_as_default_value: bool = True
    is_mandatory: bool = True
    tooltip: str | None = None

    # return in the same format that sys.file stores it in
    def metadata(self):
        base: Dict[str, Any] = {"Name": self.name, "Type": self.type.value, "Description": self.tooltip}
        if self.set_as_default_value and self.type != ParameterType.Table:
            base["DefaultValue"] = self.value
        if self.type == ParameterType.Table:
            if self.is_mandatory:
                base["ConditionUsage"] = 2
            else:
                base["ConditionUsage"] = 0
        return base


def lumi_fmt(value):
    if isinstance(value, Variable):
        return f"{value.type.value}{value.name}"
    if isinstance(value, bool):
        return "true" if value else "false"
    if isinstance(value, (int, float)):
        return str(value)
    return '"' + value.replace('"', '""') + '"'


class ViewRef(BaseModel, Ref):
    """Reference an existing view


    Example
    ----------
    >>> from fbnconfig import lumi
    >>> lumi.ViewRef(
    ...  id="lumi-example-ref",
    ...  provider="Views.fbnconfig.existing_view")


    Attributes
    ----------
    id : str
         Resource identifier.
    provider : str
        Name of the view referenced. This is assumed to exist
    """

    id: str = Field(exclude=True)
    provider: str
    _backoff_handler = BackoffHandler()
    _version: str | None = None

    def attach(self, client) -> None:
        res = query(
            client,
            textwrap.dedent(f"""\
                select r.Version from sys.registration as r
                where name = '{self.provider}'
                order by r.Version asc
                limit 1
            """),
            backoff_handler=self._backoff_handler,
        )
        if len(res) != 1:
            raise RuntimeError(f"Failed to attach ref to {self.provider}, the view might not exist")
        self._version = res[0]


class ViewResource(BaseModel, Resource):
    """Create and manage a Luminesce view

    Example
    ----------
    >>> from fbnconfig import lumi
    >>> lumi.ViewResource(
    ...  id="lumi-example-view",
    ...  provider="Views.fbnconfig.example",
    ...  description="My resource test view",
    ...  documentation_link="http://example.com/query",
    ...  variable_shape=False,
    ...  use_dry_run=True,
    ...  allow_execute_indirectly=False,
    ...  distinct=True,
    ...  sql='select 2+#PARAMETERVALUE(p1)   as   twelve',
    ...  parameters=[
    ...      lumi.Parameter(
    ...          name="p1",
    ...          value=10,
    ...          set_as_default_value=True,
    ...          tooltip="a number",
    ...          type=lumi.ParameterType.INT
    ...      )

    Attributes
    ----------
    id : str
         Resource identifier.
    provider : str
        Name of the view managed by this resource
    description : str
        View description
    sql: str
        The query string for the view
    parameters : list of `Parameter`, optional
        List of parameters for the view
    dependencies : list of dependencies, optional
        This can be another view or any other resource
    documentation_link: str, optional
        Displays one or more hyperlinks in the summary dialog for the view
    variable_shape: bool, optional
        This is useful if data returned is likely to vary in shape between queries. Defaults to false.
    allow_execute_indirectly : bool, optional
        Allows end users to query providers within the view even if they are not entitled to use those
        providers directly.
        Defaults to false.
    limit: int, optional
        Test option when developing view, does not have an effect on a published view. Defaults to None
    group_by: str, optional
        Test option when developing view, does not have an effect on a published view. Defaults to None
    filter: str, optional
        Test option when developing view, does not have an effect on a published view. Defaults to None
    offset: int, optional
        Test option when developing view, does not have an effect on a published view. Defaults to None
    distinct: bool, optional
        Test option when developing view, does not have an effect on a published view. Defaults to None
    use_dry_run: bool, optional
        Intended for automatic deployment of views. See docs for more details. Defaults to false
    variables: List of `Variable`, optional
        A table variable that can be passed into the view by an end user or in code

    See Also
    --------
    `https://support.lusid.com/knowledgebase/article/KA-01767/en-us`__
    """

    id: str = Field(exclude=True)
    provider: str = Field(serialization_alias="Provider")
    description: str = Field(serialization_alias="Description")
    sql: str
    parameters: List[Parameter] = []
    dependencies: List | None = None
    documentation_link: str | None = None
    variable_shape: bool | None = None
    allow_execute_indirectly: bool | None = None
    limit: int | None = None
    group_by: str | None = None
    filter: str | None = None
    offset: int | None = None
    distinct: bool | None = None
    use_dry_run: bool | None = None
    variables: List[Variable] = []

    _backoff_handler = BackoffHandler()

    class Registration:
        tries = 10
        wait_time = 1

    _saved_options = {  # maps from sys.file metadata to view option names
        "Description": "Description",
        "DocumentationLink": "documentationLink",
        "IsWithinDirectProviderView": "variableShape",
        "IsWithinViewAllowingIndirectExecute": "allowExecuteIndirectly",
    }
    _test_options = ["distinct", "filter", "groupby", "limit", "offset", "preamble", "useDryRun"]

    def read(self, client, old_state) -> Dict[str, Any]:
        path = old_state.provider.replace(".", "/")
        res = query(
            client,
            textwrap.dedent(f"""\
                select f.Content, r.Version from sys.file as f
                join sys.registration as r on r.Name = '{old_state.provider}'
                where path = 'databaseproviders/{path}.sql'
                order by r.Version asc
                limit 1
            """),
            self._backoff_handler,
        )
        assert len(res) == 1

        def strip_column_description(kv: Dict) -> Dict:
            if kv["Type"] == "Table":
                kv["Description"] = kv["Description"].split("\nAvailable columns")[0]

            return kv

        parts = res[0]["Content"].split("--- MetaData ---")
        sql = parts[0]
        metadata = json.loads(parts[1])

        parameters = [strip_column_description(p) for p in metadata["Parameters"]]
        props = {
            v: metadata[k] for k, v in self._saved_options.items() if metadata.get(k, None) is not None
        }
        return {"sql": sql, "version": res[0]["Version"], "parameters": parameters} | props

    @staticmethod
    def registration_version(client, view_name, backoff_handler: BackoffHandler) -> int | None:
        content = textwrap.dedent(f"""\
            select Version from sys.registration where Name='{view_name}'
            order by Version asc
            limit 1
        """)
        rows = query(client, content, backoff_handler)
        return int(rows[0]["Version"]) if len(rows) > 0 else None

    @staticmethod
    def format_option(option, value):
        if isinstance(value, bool) and value:
            return f"--{option}"
        if isinstance(value, (int, float)):
            return f"--{option}={value}"
        # we run self.dump by alias, these two have a serialization alias on the dto
        # which ends up as the option, which should be lower
        if option == "Provider" or option == "Description":
            option = option.lower()
        return f"--{option}={lumi_fmt(value)}"

    def get_variables(self):
        param_variables = [param.value for param in self.parameters if isinstance(param.value, Variable)]
        seen = set()
        return [
            value
            for value in self.variables + param_variables
            if value.name not in seen and not seen.add(value.name)
        ]

    def template(self, desired):
        options = [
            self.format_option(option, desired[option])
            for option in ["Provider"] + self._test_options + list(self._saved_options.values())
            if desired.get(option) is not None
        ]

        tpl = textwrap.dedent("""\
            {preamble}@x = use Sys.Admin.SetupView{with_clause}
            {options}{params}
            ----
            {sql}
            enduse;
            select * from @x;
        """)
        params = [
            f"{p.name},{p.type.value},{lumi_fmt(p.value)},{lumi_fmt(p.is_mandatory)}"
            + (f',"{p.tooltip}"' if p.tooltip is not None else "")
            if p.type == ParameterType.Table
            else f"{p.name},{p.type.value},{lumi_fmt(p.value)},{lumi_fmt(p.set_as_default_value)}"
            + (f',"{p.tooltip}"' if p.tooltip is not None else "")
            for p in self.parameters
        ]
        param_clause = "\n--parameters\n{0}".format("\n".join(params)) if len(params) > 0 else ""
        variables = self.get_variables()
        preamble = ";\n".join([v.init_str() for v in variables]) + ";\n" if len(variables) > 0 else ""
        with_clause = (
            " with " + ", ".join([v.with_str() for v in variables]) if len(variables) > 0 else ""
        )
        sql = tpl.format(
            options="\n".join(options),
            params=param_clause,
            sql=desired["sql"],
            with_clause=with_clause,
            preamble=preamble,
        )
        return sql

    def _get_content_hash(self) -> str:
        desired = self.model_dump(exclude_none=True, by_alias=True)
        sql = self.template(desired)
        return hashlib.sha256(sql.encode()).hexdigest()

    def create(self, client) -> Dict[str, Any]:
        desired = self.model_dump(exclude_none=True, by_alias=True)
        sql = self.template(desired)
        query(client, sql, self._backoff_handler)
        for i in range(1, self.Registration.tries + 1):
            if self.registration_version(client, self.provider, self._backoff_handler) is not None:
                break
            else:
                if i == self.Registration.tries:
                    sys.stderr.write(
                        f"warning: no view registration after {i} tries for {self.provider}"
                    )
                else:
                    time.sleep(self.Registration.wait_time)
        return {"provider": self.provider}

    def update(self, client, old_state):
        if self.provider != old_state.provider:
            self.delete(client, old_state)
            self.create(client)
            return {"provider": self.provider}
        desired = self.model_dump(exclude_none=True, by_alias=True, exclude=set(self._test_options))
        raw_remote = self.read(client, old_state)
        remote_props = {
            k.lower(): v
            for k, v in raw_remote.items()
            if k in self._saved_options.values() and v is not None
        }
        desired_props = {
            k.lower(): v
            for k, v in desired.items()
            if k in self._saved_options.values() and v is not None
        }
        remote_params = raw_remote["parameters"]
        desired_params = [p.metadata() for p in self.parameters]
        effective_params = [
            (remote_params[i] | desired_params[i])
            if i < len(remote_params) and remote_params[i]["Name"] == desired_params[i]["Name"]
            else desired_params[i]
            for i, des in enumerate(desired_params)
        ]
        remote_sql = textwrap.dedent(raw_remote["sql"].rstrip())
        remote_version = raw_remote["version"]
        desired_sql = textwrap.dedent(self.sql.rstrip())
        if (
            desired_sql == remote_sql
            and remote_props | desired_props == remote_props
            and effective_params == remote_params
        ):
            return None
        sql = self.template(desired)
        query(client, sql, self._backoff_handler)
        for i in range(1, self.Registration.tries + 1):
            version: int | None = self.registration_version(client, self.provider, self._backoff_handler)
            if version is not None and remote_version < version:
                break
            else:
                if i == self.Registration.tries:
                    sys.stderr.write(
                        f"warning: no view registration after {i} tries for {self.provider}"
                    )
                else:
                    time.sleep(self.Registration.wait_time)
        return {"provider": self.provider}

    def deps(self):
        return self.dependencies if self.dependencies else []

    @staticmethod
    def delete(client, old_state):
        sql = textwrap.dedent(f"""\
        @x = use Sys.Admin.SetupView
        --provider={old_state.provider}
        --deleteProvider
        ----
        select 1 as deleting
        enduse;
        select * from @x;
        """)
        backoff = BackoffHandler()
        query(client, textwrap.dedent(sql), backoff_handler=backoff)
        for i in range(1, ViewResource.Registration.tries + 1):
            if (
                ViewResource.registration_version(
                    client=client, view_name=old_state.provider, backoff_handler=backoff
                )
                is None
            ):
                break
            else:
                if i == ViewResource.Registration.tries:
                    sys.stderr.write(
                        f"warning: no view deregistration after {i} tries for {old_state.provider}"
                    )
                else:
                    time.sleep(ViewResource.Registration.wait_time)



from __future__ import annotations

from enum import Enum
from types import SimpleNamespace
from typing import Any, Dict, List

import httpx
from pydantic import BaseModel, ConfigDict, Field, field_serializer, model_validator

from .datatype import DataTypeRef, DataTypeResource
from .resource_abc import CamelAlias, Ref, Resource


class LifeTime(str, Enum):
    Perpetual = "Perpetual"
    TimeVariant = "TimeVariant"


class Domain(str, Enum):
    Abor = "Abor"
    AborConfiguration = "AborConfiguration"
    AccessMetadata = "AccessMetadata"
    Account = "Account"
    Allocation = "Allocation"
    Analytic = "Analytic"
    Block = "Block"
    Calendar = "Calendar"
    ChartOfAccounts = "ChartOfAccounts"
    Compliance = "Compliance"
    ConfigurationRecipe = "ConfigurationRecipe"
    CustodianAccount = "CustodianAccount"
    CustomEntity = "CustomEntity"
    CutLabelDefinition = "CutLabelDefinition"
    DerivedValuation = "DerivedValuation"
    DiaryEntry = "DiaryEntry"
    Execution = "Execution"
    Fund = "Fund"
    Holding = "Holding"
    Instrument = "Instrument"
    InstrumentEvent = "InstrumentEvent"
    Leg = "Leg"
    LegalEntity = "LegalEntity"
    MarketData = "MarketData"
    NextBestAction = "NextBestAction"
    NotDefined = "NotDefined"
    Order = "Order"
    OrderInstruction = "OrderInstruction"
    Package = "Package"
    Participation = "Participation"
    Person = "Person"
    Placement = "Placement"
    Portfolio = "Portfolio"
    PortfolioGroup = "PortfolioGroup"
    PropertyDefinition = "PropertyDefinition"
    Reconciliation = "Reconciliation"
    ReferenceHolding = "ReferenceHolding"
    Transaction = "Transaction"
    TransactionConfiguration = "TransactionConfiguration"
    UnitResult = "UnitResult"


class ResourceId(BaseModel):
    scope: str
    code: str


class ConstraintStyle(str, Enum):
    Property = "Property"
    Collection = "Collection"
    Identifier = "Identifier"


class CollectionType(str, Enum):
    Set = "Set"
    Array = "Array"


class DefinitionRef(BaseModel, Ref):
    id: str = Field(exclude=True)
    domain: Domain
    scope: str
    code: str

    def __format__(self, _):
        return f"Properties[{self.domain.value}/{self.scope}/{self.code}]"

    def attach(self, client):
        domain, scope, code = self.domain.value, self.scope, self.code
        try:
            client.get(f"/api/api/propertydefinitions/{domain}/{scope}/{code}")
        except httpx.HTTPStatusError as ex:
            if ex.response.status_code == 404:
                raise RuntimeError(f"Property definition {domain}/{scope}/{code} not found")
            else:
                raise ex


class Formula:
    formula: str
    args: Dict

    def __init__(self, formula, **kwargs):
        self.formula = formula
        self.args = kwargs


class DefinitionResource(CamelAlias, BaseModel, Resource):
    id: str = Field(init=True, exclude=True)
    domain: Domain
    scope: str
    code: str
    display_name: str
    data_type_id: ResourceId | DataTypeRef | DataTypeResource
    property_description: str | None = None
    life_time: LifeTime | None = None
    constraint_style: ConstraintStyle | None = None
    collection_type: str | None = None
    derivation_formula: Formula | None = None
    remote: Dict[str, Any] | None = Field(None, exclude=True, init=False)
    model_config = ConfigDict(arbitrary_types_allowed=True)

    def __format__(self, _):
        return f"Properties[{self.domain.value}/{self.scope}/{self.code}]"

    @field_serializer("derivation_formula", when_used="always")
    def formula_ser(self, value: Formula) -> str:
        # uses str.format to interpolate the values
        return value.formula.format(**value.args)

    @field_serializer("data_type_id", when_used="always")
    def datatype_ser(self, value: ResourceId | DataTypeRef | DataTypeResource) -> Dict[str, str]:
        # uses str.format to interpolate the values
        return {"scope": value.scope, "code": value.code}

    @model_validator(mode="after")
    def check_derived_or_plain(self):
        if self.derivation_formula:
            if self.life_time or self.constraint_style or self.collection_type:
                raise RuntimeError("A property must be either derived or plain")
        return self

    def read(self, client, old_state):
        domain = old_state.domain
        scope = old_state.scope
        code = old_state.code
        self.remote = client.get(f"/api/api/propertydefinitions/{domain}/{scope}/{code}").json()
        return self.remote

    def create(self, client: httpx.Client):
        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True)
        derived = self.derivation_formula is not None
        if derived:
            client.request("POST", "/api/api/propertydefinitions/derived", json=desired)
        else:
            client.request("POST", "/api/api/propertydefinitions", json=desired)
        return {"domain": self.domain.value, "scope": self.scope, "code": self.code, "derived": derived}

    @staticmethod
    def delete(client, old_state: SimpleNamespace):
        domain, scope, code = old_state.domain, old_state.scope, old_state.code
        client.delete(f"/api/api/propertydefinitions/{domain}/{scope}/{code}")

    def update(self, client, old_state):
        # cannot change identifier or switch derived and non-derived. recreate
        derived = self.derivation_formula is not None
        if [self.domain.value, self.scope, self.code, derived] != [
            old_state.domain,
            old_state.scope,
            old_state.code,
            old_state.derived,
        ]:
            self.delete(client, old_state)
            return self.create(client)
        self.read(client, old_state)
        remote = self.remote or {}
        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True)
        effective = remote | desired
        if effective == remote:
            return None
        # cannot change dataType, or collectionType need to recreate
        if remote["dataTypeId"] != desired["dataTypeId"] or remote.get(
            "collectionType", None
        ) != effective.get("collectionType", None):
            self.delete(client, old_state)
            return self.create(client)
        if derived:
            client.put(
                f"/api/api/propertydefinitions/derived/{self.domain.value}/{self.scope}/{self.code}",
                json=desired,
            )
        else:
            client.put(
                f"/api/api/propertydefinitions/{self.domain.value}/{self.scope}/{self.code}",
                json=desired,
            )
        return {"domain": self.domain.value, "scope": self.scope, "code": self.code, "derived": derived}

    def deps(self):
        res: List[Resource | Ref] = []
        if self.derivation_formula:
            res = [
                value
                for value in self.derivation_formula.args.values()
                if isinstance(value, (DefinitionResource, DefinitionRef))
            ]
        if isinstance(self.data_type_id, (Resource, Ref)):
            res.append(self.data_type_id)
        return res
